@@ -6,7 +6,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2024	PgPool Global Development Group
+ * Copyright (c) 2003-2023	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -160,6 +160,12 @@ typedef struct {
 }			POOL_TEMP_TABLE;
 
 
+typedef enum
+{
+	SI_NO_SNAPSHOT,
+	SI_SNAPSHOT_PREPARED
+} SI_STATE;
+
 /*
  * Per session context:
  */
@@ -282,25 +288,27 @@ typedef struct
 	 */
 	List	   *temp_tables;
 
+	bool		is_in_transaction;
+
+	/*
+	 * Current transaction temp write list
+	 */
+	List	   *transaction_temp_write_list;
+
 #ifdef NOT_USED
-	/* Preferred "master" node id. Only used for SimpleForwardToFrontend. */
-	int			preferred_master_node_id;
+	/* Preferred "main" node id. Only used for SimpleForwardToFrontend. */
+	int			preferred_main_node_id;
 #endif
 
-       bool            is_tx_started_by_multi_statement;       /* True if an explicit
-																* transaction has been
-																* started by a
-																* multi-statement-query */
-	/*
-	 * True if query cache feature disabled until session ends.
-	 * This is set when SET ROLE/SET SESSION AUTHORIZATION executed.
-	 */
-	bool	query_cache_disabled;
-	/*
-	 * True if query cache feature disabled until current transaction ends.
-	 * This is set when REVOKE executed in a transaction.
-	 */
-	bool	query_cache_disabled_tx;
+	/* Whether snapshot is aquired in this transaction. Only used by Snapshot Isolation mode. */
+	SI_STATE	si_state;
+	/* Whether transaction is read only. Only used by Snapshot Isolation mode. */
+	SI_STATE	transaction_read_only;
+
+	bool		is_tx_started_by_multi_statement;	/* True if an explicit
+													 * transaction has been
+													 * started by a
+													 * multi-statement-query */
 }			POOL_SESSION_CONTEXT;
 
 extern void pool_init_session_context(POOL_CONNECTION * frontend, POOL_CONNECTION_POOL * backend);
@@ -388,11 +396,15 @@ extern bool is_tx_started_by_multi_statement_query(void);
 extern void set_tx_started_by_multi_statement_query(void);
 extern void unset_tx_started_by_multi_statement_query(void);
 
-extern void set_query_cache_disabled(void);
-extern void unset_query_cache_disabled(void);
-extern bool query_cache_disabled(void);
-extern void set_query_cache_disabled_tx(void);
-extern void unset_query_cache_disabled_tx(void);
-extern bool query_cache_disabled_tx(void);
+#ifdef NOT_USED
+extern void pool_set_preferred_main_node_id(int node_id);
+extern int	pool_get_preferred_main_node_id(void);
+extern void pool_reset_preferred_main_node_id(void);
+#endif
 
+#ifdef NOT_USED
+extern void pool_add_prep_where(char *name, bool *map);
+extern bool *pool_get_prep_where(char *name);
+extern void pool_delete_prep_where(char *name);
+#endif							/* NOT_USED */
 #endif							/* POOL_SESSION_CONTEXT_H */
